#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <pwd.h>
#include <signal.h>
#include <time.h>
#include <errno.h>
#include "sock-priv.h"
#ifndef TWOPLAYER
#define TWOPLAYER
#endif
#ifndef SOCKET
#define SOCKET
#endif
#include "tetris.h"
#include "socket.h"
#include "input.h"
#include "draw.h"

const struct invit *invit = NULL;
int sock_flags = -1;

struct sockaddr *sockaddr = NULL;

static char this_tty[8] = "";
static int invit_fd = -1;

static int get_this_tty()
{
	char *s = ttyname(STDOUT_FILENO);
	int i;
	if (!s)
		return 0;
	if (!strncmp(s, "/dev/", 5))
		i = 5;
	else if (strlen(s) < 8)
		i = 0;
	else {
		i = strlen(s)-7;
		if (s[i] == '/')
			i++;
	}
	strcpy(this_tty, s+i);
	return 1;
}

static void getfname(char *name, const char *tty)
{
	int i;
	memcpy(name, "/tmp/vitetris-", 14);
	strncpy(name+14, tty, 7);
	name[21] = '\0';
	for (i = 14; i < 20 && name[i]; i++)
		if (name[i] == '/')
			name[i] = '-';
}

void mkinvitfile()
{
	char name[22];
	if (*this_tty || get_this_tty()) {
		getfname(name, this_tty);
		invit_fd = open(name, O_CREAT | O_RDONLY | O_NONBLOCK);
		chmod(name, S_IREAD | S_IWRITE | S_IWGRP | S_IWOTH);
	}
}

static int getowner(const char *tty, char *username)
{
	char fname[22];
	struct stat st;
	struct passwd *pw;
	getfname(fname, tty);
	if (stat(fname, &st) == -1)
		return 0;
	if (pw = getpwuid(st.st_uid)) {
		strncpy(username, pw->pw_name, 15);
		username[15] = '\0';
		return 1;
	}
	return 0;
}

int checkinvit()
{
	static struct invit inv;
	int i = 0;
	ssize_t n;
	invit = NULL;
	if (invit_fd == -1)
		return 0;
	while (i < 8) {
		n = read(invit_fd, inv.tty+i, 1);
		if (n == -1 && (errno==EAGAIN || errno==EINTR))
			continue;
		if (n < 1)
			break;
		if (inv.tty[i] == '\n') {
			inv.tty[i] = '\0';
			break;
		}
		i++;
	}
	if (i==8 || !i && !n && lseek(invit_fd, 0, SEEK_CUR) > 0) {
		rminvitfile();
		mkinvitfile();
		i = 0;
	}
	if (i && getowner(inv.tty, inv.user)) {
		invit = &inv;
		return 1;
	}
	return 0;
}

void rminvitfile()
{
	char name[22];
	if (invit_fd > -1) {
		close(invit_fd);
		invit_fd = -1;
		getfname(name, this_tty);
		unlink(name);
	}
}

static int add_2p_tty(char *tty, const char *name)
{
	struct stat st;
	char fname[22];
	int i;
	getfname(fname, name);
	if (!islower(*name) || strchr(name, '.') || stat(fname, &st) < 0)
		return 0;
	if (time(NULL) - st.st_mtime > 3600) {
		unlink(fname);
		return 0;
	}
	strncpy(tty, name, 7);
	for (i = 1; i < 8 && tty[i]; i++)
		if (tty[i] == '-')
			tty[i] = '/';
	return strcmp(tty, this_tty) != 0;
}

int get_2p_ttys(char *ttys, int n)
{
	DIR *d = opendir("/tmp");
	struct dirent *e;
	int i;
	if (!d)
		return 0;
	i = 0;
	while (i < n && (e = readdir(d))) {
		if (strncmp(e->d_name, "vitetris-", 9))
			continue;
		if (add_2p_tty(ttys+8*i, e->d_name+9))
			i++;
	}
	if (i && i < 3)
		ttys[8*i] = '\0';
	closedir(d);
	return i > 0;
}

static char *errmsg(const char *descr)
{
	char *msg, *s;
	int n = strlen(descr);
	s = strerror(errno);
	msg = malloc(8+n+(s ? strlen(s)+2 : 0));
	memcpy(msg, "ERROR! ", 7);
	strcpy(msg+7, descr);
	if (s) {
		n += 7;
		msg[n] = ':';
		msg[n+1] = ' ';
		strcpy(msg+n+2, s);
	}
	return msg;
}

static void conn_broken(int sig)
{
	sock_flags |= CONN_BROKEN;
	sock_flags &= ~PL2_IN_GAME;
}

char *mksocket(int domain)
{
	int sock = socket(domain, SOCK_STREAM, 0);
	if (sock < 0)
		return errmsg("Could not create socket");
	socket_fd = sock;
	if (sock_flags == -1)
		signal(SIGPIPE, conn_broken);
	sock_flags = 0;
	return NULL;
}

static void getaddr(const char *tty, int server)
{
	char fname[29];
	struct sockaddr_un *addr;
	if (server) {
		tty = this_tty;
		if (!*tty)
			get_this_tty();
	}
	getfname(fname, tty);
	strcat(fname+14, ".socket");
	if (sockaddr)
		free(sockaddr);
	addr = malloc(offsetof(struct sockaddr_un, sun_path)
					+ strlen(fname)+1);
	addr->sun_family = AF_UNIX;
	strcpy(addr->sun_path, fname);
	sockaddr = (struct sockaddr *) addr;
}

char *bind_listen(int size)
{
	int sock = socket_fd;
	if (bind(sock, sockaddr, size) == -1)
		return errmsg("Could not bind socket");
	if (listen(sock, 1) == -1)
		return errmsg("listen");
	sock_flags |= IS_SERVER;
	return NULL;
}

static char *inviteplayer(const char *tty)
{
	char name[22];
	int fd;
	int n = strlen(this_tty);
	getfname(name, tty);
	fd = open(name, O_WRONLY | O_APPEND);
	if (fd < 0)
		return errmsg(name);
	this_tty[n] = '\n';
	while (write(fd, this_tty, n+1) == -1 && errno==EINTR)
		;
	close(fd);
	this_tty[n] = '\0';
	return NULL;
}

int connectsock(int size)
{
	if (connect(socket_fd, sockaddr, size) == -1)
		return 0;
	sock_flags |= CONNECTED;
	return 1;
}

static char *errmsg_connect_local()
{
	char *msg;
	char *s = malloc(strlen(SUN_PATH(sockaddr))+9);
	strcpy(s, "connect ");
	strcat(s, SUN_PATH(sockaddr));
	msg = errmsg(s);
	free(s);
	return msg;
}

char *mksocket_local(const char *tty, int server)
{
	char *msg = mksocket(PF_UNIX);
	int size;
	if (!msg) {
		getaddr(tty, server);
		size = offsetof(struct sockaddr_un, sun_path)
				  + strlen(SUN_PATH(sockaddr));
		if (server) {
			msg = bind_listen(size);
			chmod(SUN_PATH(sockaddr), S_IREAD | S_IWRITE |
			      S_IRGRP | S_IROTH | S_IWGRP | S_IWOTH);
			if (!msg && tty)
				msg = inviteplayer(tty);
		} else if (!connectsock(size))
			msg = errmsg_connect_local();
		if (msg)
			rmsocket();
	}
	return msg;
}

void connect_tty(const char *tty)
{
	char *msg, *p;
	if (!strncmp(tty, "/dev/", 5))
		tty += 5;
	printf("Connecting to %s...\n", tty);
	msg = mksocket_local(tty, 0);
	if (!msg)
		return;
	p = strchr(msg, '/');
	puts(p);
	free(msg);
	if (!get_this_tty())
		exit(1);
	if (!strncmp(tty, this_tty, 7)) {
		printf("%s is this terminal!\n", tty);
		exit(1);
	}
	puts("Failed.");
	printf("Sending invitation to %s...\n", tty);
	mkinvitfile();
	if (msg = mksocket_local(tty, 1)) {
		puts(msg);
		free(msg);
		exit(1);
	}
}

static void accept_conn(int fd)
{
	fd = accept(fd, NULL, NULL);
	rmsocket();
	socket_fd = fd;
	sock_flags |= CONNECTED;
}

static void writebytes(int fd, const char *buf, int n)
{
	int i = write(fd, buf, n);
	if (i == n)
		return;
	if (i == -1) {
		if (errno != EINTR)
			return;
		i = 0;
	}
	writebytes(fd, buf+i, n-i);
}

static void writeb(int fd, char b)
{
	writebytes(fd, &b, 1);
}

void sock_sendbyte(char b)
{
	writeb(socket_fd, b);
}

static int waitinput_1sec(int fd)
{
	int t = gettm(0);
	while (!waitinput(fd, 50000))
		if (gettm(t)-t > 950)
			return 0;
	return 1;
}

static int readbytes(int fd, char *buf, int n)
{
	int i;
	if (sock_flags & CONN_BROKEN)
		return 0;
	if (!waitinput_1sec(fd))
		goto broken;
	i = read(fd, buf, n);
	if (i == n)
		return 1;
	if (i == -1 && errno == EINTR)
		i = 0;
	else if (i <= 0) {
broken:		conn_broken(0);
		return 0;
	}
	return readbytes(fd, buf+i, n-i);
}

void sock_initgame()
{
	int fd = socket_fd;
	if (is_server)
		player2.lineslimit = player1.lineslimit;
	else {
		writeb(fd, 'i');
		sock_sendplayer();
	}
}

void sock_sendplayer()
{
	char s[3] = "p";
	s[1] = player1.startlevel;
	s[2] = player1.rotationsys;
	writebytes(socket_fd, s, 3);
}

static void recvplayer(int fd)
{
	char s[2];
	if (readbytes(fd, s, 2)) {
		if (s[0] > 9 || s[0] < 0) {
			sock_flags |= CONN_BROKEN;
			return;
		}
		player2.startlevel = s[0];
		player2.rotationsys = s[1];
	}
}

static int getkey_cancel()
{
	switch (getkeypress(500, SINGLE_PL)) {
	case ESC:
	case '\b':
		return 1;
	case 'q':
		exit(0);
	}
	return sock_flags & CONN_BROKEN;
}

int sock_wait_pl2ingame()
{
	while (!(sock_flags & CONNECTED))
		if (getkey_cancel())
			return 0;
	writeb(socket_fd, 'G');
	sock_flags |= WAIT_PL2INGAME;
	while (!(sock_flags & PL2_IN_GAME))
		if (getkey_cancel())
			break;
	sock_flags &= ~WAIT_PL2INGAME;
	return sock_flags & PL2_IN_GAME;
}

static void sendboard(int fd, int n)
{
	char s[4] = "b";
	unsigned char *b;
	uint_least32_t row;
	int i = 20;
	if (!(sock_flags & PL2_IN_GAME))
		return;
	s[1] = n;
	writebytes(fd, s, 2);
	while (n) {
		memset(s, 0, 4);
		row = player1.board[--i];
		b = (unsigned char *) s;
		do {
			*b = row & 0xFF;
			b++;
		} while (row >>= 8);
		writebytes(fd, s, 4);
		n--;
	}
}

static void recvboard(int fd)
{
	char s[48];
	unsigned char *b = (unsigned char *) s;
	uint_least32_t row;
	int n;
	int i = 20;
	int j;
	if (!readbytes(fd, s, 1)) 
		return;
	n = *s;
	if (n < 1 || n > 12) {
		conn_broken(0);
		return;
	}
	if (!readbytes(fd, s, 4*n))
		return;
	while (n) {
		row = b[3];
		for (j = 2; j >= 0; j--) {
			row <<= 8;
			row |= b[j];
		}
		player2.board[--i] = row;
		b += 4;
		n--;
	}
	redrawboard(&player2, 19);
}

static void sendboard_height()
{
	int n = player1.height;
	switch (n) {
	case 1:
	case 2:
		n = 2*n+1;
		break;
	default:
		n = 2*(n+1);
	}
	sendboard(socket_fd, n);
}

void sock_initround()
{
	char s[3] = "n1";
	int fd = socket_fd;
	writeb(fd, 'G');
	if (is_server) {
		s[2] = *tetrom_seq;
		writebytes(fd, s, 3);
		s[1] = '2';
		writebytes(fd, s, 3);
		s[2] = tetrom_seq[1];
		writebytes(fd, s, 3);
		s[1] = '1';
		writebytes(fd, s, 3);
	}
	if (player1.height)
		sendboard_height();
}

void sock_sendnext(const struct player *p, char n)
{
	char s[3] = "n";
	if (sock_flags & PL2_IN_GAME) {
		s[1] = '1'+!isplayer2(p);
		s[2] = n;
		writebytes(socket_fd, s, 3);
	}
}

static void recvnext(int fd)
{
	struct tetr next;
	char s[2];
	char *p = tetrom_seq;
	if (!p || !readbytes(fd, s, 2) || s[1] > 6)
		return;
	switch (*s) {
	case '2':
		p += 2;
	case '1':
		break;
	default:
		return;
	}
	if (*p >= 0)
		p++;
	else {
		gettetrom(&next, s[1]);
		drawnext(game->player+(*s=='2'), &next);
		if (game->next)
		    *game->next = next;
	}
	*p = s[1];
}

void sock_sendpiece(const struct player *p)
{
	char s[5] = "x";
	if (sock_flags & PL2_IN_GAME) {
		int bl = player1.piece.blocks;
		s[1] = bl & 0xFF;
		s[2] = bl>>8;
		memcpy(s+3, &player1.piece.x, 2);
		writebytes(socket_fd, s, 5);
	}
}

static void recvpiece(int fd)
{
	char bl[2];
	if (readbytes(fd, bl, 2)) {
		player2.piece.blocks = bl[0] | bl[1]<<8;
		readbytes(fd, &player2.piece.x, 2);
	}
}

void sock_requestgarbage()
{
	char s[2] = "g";
	int n = 1;
	while (!player2.board[19-n] && n < 12)
		n++;
	s[1] = n;
	writebytes(socket_fd, s, 2);
}

int sock_getkeypress(int flags)
{
	int fd = socket_fd;
	char b;
	if (!(sock_flags & CONNECTED)) {
		accept_conn(fd);
		return 0;
	}
	if ((sock_flags & CONN_BROKEN) && !(flags & IN_GAME) && !game_over)
		return ESC;
	if (!readbytes(fd, &b, 1))
		return 0;
	if (b=='G' && game_over) {
		sock_flags &= ~PL2_IN_GAME;
		writeb(fd, ESC);
		goto out;
	}
	switch (b) {
	case ESC:
		sock_flags &= ~PL2_IN_GAME;
		break;
	case 'G':
		if (!(sock_flags & WAIT_PL2INGAME))
			break;
		sock_flags |= PL2_IN_GAME;
	case '.':
		return '.';
	case 'b':
		if (tetrom_seq) {
			recvboard(fd);
			if (!game_running)
				print_press_key();
		}
		break;
	case 'g':
		if (readbytes(fd, &b, 1) && b > 0 && b <= 12)
			sendboard(fd, b);
		break;
	case 'h':
		if (readbytes(fd, &b, 1)) {
			if (isdigit(b))
				player2.height = b-'0';
			else {
				writeb(fd, 'h');
				writeb(fd, player1.height+'0');
			}
		}
		break;
	case 'i':
		sock_sendplayer();
		writeb(fd, 'm');
		writeb(fd, game->mode);
		writeb(fd, player1.lineslimit);
		break;
	case 'm':
		readbytes(fd, &game->mode, 1);
		readbytes(fd, &player1.lineslimit, 1);
		player2.lineslimit = player1.lineslimit;
		break;
	case 'n':
		recvnext(fd);
		break;
	case 'p':
		sock_flags &= ~PL2_IN_GAME;
		recvplayer(fd);
		break;
	case 's':
		if (game_over)
			return '\n';
		break;
	case 'x':
		recvpiece(fd);
		break;
	default:
		if (flags & IN_GAME && (b > 0 && b <= B_BTN ||
					b == HARDDROP))
			return b;
	}
out:	if (waitinput(fd, 0))
		return sock_getkeypress(flags);
	return 0;
}

void rmsocket()
{
	if (sockaddr) {
		if (is_server && sockaddr->sa_family == AF_UNIX)
			unlink(SUN_PATH(sockaddr));
		free(sockaddr);
		sockaddr = NULL;
	}
	if (socket_fd > -1) {
		close(socket_fd);
		socket_fd = -1;
	}
}
