/* term flags */
#define ASCII 1
#define WHITE_BG 2
#define MONOCHROME 4
#define HEIGHT_24L 8
#define TT_BLOCKS 16
#define TT_BLOCKS_BG 32
#define XTERM 64
#define LINUX_TERM 0x80
#define GNOME_TERM 0x100
#define CYGWIN 0x200
/* TERM_RESIZED 0x400 defined in xterm.h */

#define _ASCII (term_flags & ASCII)
#define _WHITE_BG (term_flags & WHITE_BG)
#define _MONOCHROME (term_flags & MONOCHROME)
#define _HEIGHT_24L (term_flags & HEIGHT_24L)
#define _TT_BLOCKS (term_flags & TT_BLOCKS)
#define TT_MONO (TT_BLOCKS | MONOCHROME)
#define _TT_BLOCKS_BG (term_flags & TT_BLOCKS_BG)
#define _XTERM (term_flags & XTERM)
#define _LINUX_TERM (term_flags & LINUX_TERM)

int term_flags;

char term_width;
char term_height;

void gettermsize();
void settermwidth(int w);
void settermheight(int h);

void term_setup();
void restoreterm();

#define WIN_NEXT 3
#define WIN_PANEL 5
#define WIN_TETROM_STATS 6
#define WIN_TOP_SCORES 7

int getmargin_x();
void getwin_xy(int win, int *x, int *y);

void setcurs(int x, int y);
void setwcurs(int win, int x, int y);
void movefwd(int n);
void newln(int x);
void setcurs_end();
int is_outside_screen(int x);
#ifndef NO_MENU
void get_xy(int *x, int *y);	/* should not be used much */
#endif

void refreshwin(int win);
void clearwin(int win);
void cleartoeol();

/* Color pairs.
 * 1-7 are used for blocks.
 * 0x11-0x17 are block colors without bg. */
#define MAGENTA_FG 8
#define WHITE_ON_BLUE 9
#define BOARD_BG_COLOR 10
#define BOARD_FRAME_COLOR 11
#define PANEL_LABEL_COLOR 12
#define RED_FG 13
#define YELLOW_ON_GREEN 14
#define YELLOW_ON_BLUE 0x18

void setcolorpair(int pair);

void setattr_normal();
void setattr_standout();
void setattr_bold();
void setattr_underline();

/* Line drawing characters - used with putch */

#define UPLEFT	 ('l' | 0x100)
#define LOWLEFT  ('m' | 0x100)
#define UPRIGHT	 ('k' | 0x100)
#define LOWRIGHT ('j' | 0x100)
#define CROSSLINES ('n' | 0x100)
#define HLINE	 ('q' | 0x100)
#define VLINE	 ('x' | 0x100)
#define LEFT_T	 ('t' | 0x100)
#define RIGHT_T	 ('u' | 0x100)
#define BOTTOM_T ('v' | 0x100)
#define TOP_T	 ('w' | 0x100)
#define TEXTURE1 ('a' | 0x100)	/* ACS_CKBOARD */
#define TEXTURE2 ('h' | 0x100)	/* ACS_BOARD or blank */
#define BULLET	 ('~' | 0x100)
#define UPARROW	 ('-' | 0x100)

void putch(int ch);
int printstr(const char *str);
void printint(const char *fmt, int d);

/* takes a line drawing character and prints ascii */
void putch_ascii(int ch);

int putnchars(int ch, int n);

/* prints lowercase letters in str using alternate character set,
 * and repeats a character followed by 'N' n times */
void printstr_acs(const char *str, int n);

int isprintable(char c);

int default_bgdot();

/* Menus */

#ifndef NO_MENU
void term_entermenu();
#ifdef TWOPLAYER
int term_startupmenu(int i);
#endif
int term_gamemenu();

void print_vitetris_ver(int x, int y);

struct termopt {
	unsigned char flag;
	const char *s;
	const char *key;
};
int term_optionhandler(int keypress, const struct termopt *o);
int term_blockstyle(int keypress);

#else /* if NO_MENU */
#define term_entermenu()
#endif

void term_entergame();
void print_top_scores();

void term_startup(int i);
