/* xterm.c - window focus change and terminal resizing */

#include <signal.h>
#ifndef XTERMWIN
#define XTERMWIN
#endif
#include "xterm.h"
#include "term.h"
#include "menu.h"	/* in_menu */
#include "tetris.h"
#include "draw.h"

#ifdef CURSES
#include <curses.h>
WINDOW *window;		/* term_curses.c */
WINDOW *wins[6];
#endif

#ifdef XLIB
#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>

int in_xterm = 0;

static Display *dpy;
static Window xtermwin;

int xterm_hasfocus()
{
	Window w, focusw, root, par, *children;
	unsigned int n;
	int i;
	if (!XGetInputFocus(dpy, &w, &i))
		return 0;
	if (w == xtermwin)
		return 1;
	focusw = w;
	w = xtermwin;
	while (XQueryTree(dpy, w, &root, &par, &children, &n) && par != root) {
		if (children) {
			for (i = 0; i < n; i++)
				if (children[i] == focusw) {
					focusw = xtermwin;
					break;
				}	
			XFree(children);
			if (focusw == xtermwin)
				return 1;
		}
		if (par == focusw)
			return 1;
		w = par;
	}
	return 0;
}

static void in_xterm_init()
{
	char *s = getenv("WINDOWID");
	if (s && sscanf(s, "%lu", &xtermwin)==1 && (dpy = XOpenDisplay(0)))
		in_xterm = xterm_hasfocus();
}
#else /* if !XLIB */
#define in_xterm_init()
#endif

static void resizehandler(int sig)
{
	term_flags |= TERM_RESIZED;
}

void xterm_init()
{
	in_xterm_init();
	signal(SIGWINCH, resizehandler);
}

static void entergame_redraw()
{
	term_entergame();
	if (!TWOPLAYER_MODE)
		drawgamescreen_1p();
	else
		drawgamescreen_2p();
	if (game->next)
	    print_press_key();
}

#if defined(CURSES) && defined(KEY_RESIZE)
static void resizemenu(int w, int h)
{
	int x, y;
	if (h > term_height) {
		getyx(window, y, x);
		wmove(window, term_height-2, 0);
		wclrtobot(window);
		wmove(window, y, x);
	}
	x = getmargin_x();
	if (w > term_width)
		mvwin(window, 1, x);
	resizeterm(term_height, term_width);
	wresize(window, term_height-1, term_width-x);
	if (w < term_width)
		mvwin(window, 1, x);
	clear();
	refresh();
	touchwin(window);
	wrefresh(window);
}

static void resizegamerunning()
{
	WINDOW *next = wins[WIN_NEXT];
	WINDOW *next2 = NULL;
	int i, x, y;
	wins[WIN_NEXT] = NULL;
	if (TWOPLAYER_MODE) {
		next2 = wins[WIN_NEXT+1];
		wins[WIN_NEXT+1] = NULL;
	}
	entergame_redraw();
	touchwin(wins[1]);
	wrefresh(wins[1]);
	if (TWOPLAYER_MODE) {
		touchwin(wins[2]);
		wrefresh(wins[2]);
	}
	if (!next)
		return;
	i = WIN_NEXT;
	while (1) {
		getwin_xy(i, &x, &y);
		x += getmargin_x();
		mvwin(next, y, x);
		overlay(next, wins[i]);
		delwin(next);
		wrefresh(wins[i]);
		if (TWOPLAYER_MODE && next != next2) {
			next = next2;
			i++;
			continue;
		}
		break;
	}
}
#endif

void upd_termresize()
{
#if !defined(CURSES) || defined(KEY_RESIZE)
	int w, h;
	if (!(term_flags & TERM_RESIZED) ||
	    !in_menu && game && game->state & (GAME_OVER
# ifndef CURSES
					| GAME_RUNNING
# endif
					))
		return;
	term_flags &= ~TERM_RESIZED;
	w = term_width;
	h = term_height;
	gettermsize();
# ifndef CURSES
	if (w == term_width && h == term_height)
		return;
	if (in_menu)
		return;
# else
	if (in_menu) {
		resizemenu(w, h);
		return;
	}
	resizeterm(term_height, term_width);
	if (game && game_running) {
		resizegamerunning();
		return;
	}
# endif
	entergame_redraw();
# ifndef CURSES
	if (term_height < 24 && term_height > 21) {
		w = -getmargin_x();
		setwcurs(0, -w, 20);
		newln(-w);
		clearbox(w, 21, 0, term_height-21);
	}
# endif
	if (game_paused)
		clearboard_paused();
	else {
		drawnext(&player1, game->next);
		if (TWOPLAYER_MODE)
			drawnext(&player2, game->next);
	}
#endif
}
